from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar

# import boto3
from s3path import S3Path
from rich.style import Style
from rich.text import Text, TextType

from textual._types import MessageTarget
from textual.message import Message
from textual.widgets._tree import TOGGLE_STYLE, Tree, TreeNode


@dataclass
class DirEntry:
    """Attaches directory information ot a node."""

    path: str
    is_dir: bool
    fs: str = "local"
    loaded: bool = False


class DirectoryFilterTree(Tree[DirEntry]):
    """A Tree widget that presents files and directories.

    Args:
        path: Path to directory.
        name: The name of the widget, or None for no name. Defaults to None.
        id: The ID of the widget in the DOM, or None for no ID. Defaults to None.
        classes: A space-separated list of classes, or None for no classes. Defaults to None.
        disabled: Whether the directory tree is disabled or not.
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "directory-tree--folder",
        "directory-tree--file",
        "directory-tree--extension",
        "directory-tree--hidden",
    }
    """
    | Class | Description |
    | :- | :- |
    | `directory-tree--extension` | Target the extension of a file name. |
    | `directory-tree--file` | Target files in the directory structure. |
    | `directory-tree--folder` | Target folders in the directory structure. |
    | `directory-tree--hidden` | Target hidden items in the directory structure. |

    See also the [component classes for `Tree`][textual.widgets.Tree.COMPONENT_CLASSES].
    """

    DEFAULT_CSS = """
    DirectoryTree > .directory-tree--folder {
        text-style: bold;
    }

    DirectoryTree > .directory-tree--file {

    }

    DirectoryTree > .directory-tree--extension {
        text-style: italic;
    }

    DirectoryTree > .directory-tree--hidden {
        color: $text 50%;
    }
    """

    class FileSelected(Message, bubble=True):
        """Posted when a file is selected.

        Can be handled using `on_directory_tree_file_selected` in a subclass of
        `DirectoryTree` or in a parent widget in the DOM.

        Attributes:
            path: The path of the file that was selected.
        """

        def __init__(self, path: str) -> None:
            self.path: str = path
            super().__init__()

    def __init__(
        self,
        path: str,
        filter: list[str],
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:

        # identity file system
        if path[0:5] == "s3://":
            self.fs = "s3"
            self.path = path[4:]  # replace s3:// with / for S3Path
        else:
            self.path = path
            self.fs = "local"

        self.filter = filter

        super().__init__(
            path,
            data=DirEntry(self.path, True, self.fs),
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

    def process_label(self, label: TextType):
        """Process a str or Text into a label. Maybe overridden in a subclass to modify how labels are rendered.

        Args:
            label: Label.

        Returns:
            A Rich Text object.
        """
        if isinstance(label, str):
            text_label = Text(label)
        else:
            text_label = label
        first_line = text_label.split()[0]
        return first_line

    def render_label(self, node: TreeNode[DirEntry], base_style: Style, style: Style):
        node_label = node._label.copy()
        node_label.stylize(style)

        if node._allow_expand:
            prefix = ("📂 " if node.is_expanded else "📁 ", base_style + TOGGLE_STYLE)
            node_label.stylize_before(
                self.get_component_rich_style("directory-tree--folder", partial=True)
            )
        else:
            prefix = (
                "📄 ",
                base_style,
            )
            node_label.stylize_before(
                self.get_component_rich_style("directory-tree--file", partial=True),
            )
            node_label.highlight_regex(
                r"\..+$",
                self.get_component_rich_style(
                    "directory-tree--extension", partial=True
                ),
            )

        if node_label.plain.startswith("."):
            node_label.stylize_before(
                self.get_component_rich_style("directory-tree--hidden")
            )

        text = Text.assemble(prefix, node_label)
        return text

    def load_directory(self, node: TreeNode[DirEntry]) -> None:
        assert node.data is not None
        node.data.loaded = True

        if self.fs == "s3":
            dir_path = S3Path(node.data.path)
        else:
            dir_path = Path(node.data.path)

        directory = sorted(
            list(dir_path.iterdir()),
            key=lambda path: (not path.is_dir(), path.name.lower()),
        )
        for path in directory:
            if path.is_dir() or path.suffix.lower() in self.filter:
                node.add(
                    path.name,
                    data=DirEntry(str(path), path.is_dir(), self.fs),
                    allow_expand=path.is_dir(),
                )
        node.expand()

    def on_mount(self) -> None:
        self.load_directory(self.root)

    def on_tree_node_expanded(self, event: Tree.NodeSelected) -> None:
        event.stop()
        dir_entry = event.node.data
        if dir_entry is None:
            return
        if dir_entry.is_dir:
            if not dir_entry.loaded:
                self.load_directory(event.node)
        else:
            self.post_message(self.FileSelected(dir_entry.path))

    def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        event.stop()
        dir_entry = event.node.data
        if dir_entry is None:
            return
        if not dir_entry.is_dir:
            if dir_entry.fs == "s3":
                path = "s3:/" + dir_entry.path
            else:
                path = dir_entry.path
            self.post_message(self.FileSelected(path))



"""
Data browser example.

Run with:

    python data_browser.py PATH

"""

import pathlib
import sys

import pandas as pd
from rich.syntax import Syntax
from rich.text import Text
from rich.traceback import Traceback
from textual import events
from textual.app import App, ComposeResult
from textual.containers import Container, Vertical
from textual.reactive import var
from textual.widgets import DataTable, Footer, Header, Static
from .widgets import DirectoryFilterTree


class Notification(Static):
    def on_mount(self) -> None:
        self.set_timer(3, self.remove)

    def on_click(self) -> None:
        self.remove()


class DataBrowser(App):
    """Textual data browser app."""

    CSS_PATH = "data_browser.css"
    BINDINGS = [
        ("f", "toggle_files", "Toggle Files"),
        ("d", "toggle_dtype", "Toggle Dtype"),
        ("s", "screenshot", "Screenshot"),
        ("q", "quit", "Quit"),
    ]

    LOADERS = {
        ".csv": pd.read_csv,
        ".parquet": pd.read_parquet,
        ".json": pd.read_json,
    }

    show_tree = var(True)
    show_dtype = var(False)

    def __load_table(self):
        self.table.clear(columns=True)
        if self.show_dtype:
            show_df = pd.DataFrame(self.df.dtypes.apply(lambda x: x.name).reset_index())
            show_df.columns = ["field", "dtype"]
        else:
            show_df = self.df

        self.table.add_columns(*show_df.columns.to_list())
        self.table.zebra_stripes = True
        for n in range(min(100, len(show_df))):
            self.table.add_row(*show_df.iloc[n])

    def watch_show_tree(self, show_tree: bool) -> None:
        """Called when show_tree is modified."""
        self.set_class(show_tree, "-show-tree")

    def compose(self) -> ComposeResult:
        """Compose our UI."""
        path = "./" if len(sys.argv) < 2 else sys.argv[1]
        yield Header()
        with Container():
            yield DirectoryFilterTree(
                path,
                self.LOADERS.keys(),
                id="tree-view",
            )
            with Vertical(id="data-view"):
                # yield Static(id="data", expand=True)
                yield DataTable(id="data")
        yield Footer()

    def on_mount(self, event: events.Mount) -> None:
        self.query_one(DirectoryFilterTree).focus()

    def on_directory_filter_tree_file_selected(
        self, event: DirectoryFilterTree.FileSelected
    ) -> None:
        """Called when the user click a file in the directory tree."""
        event.stop()
        self.table = self.query_one("#data", DataTable)

        suffix = pathlib.Path(event.path).suffix.lower()

        try:
            if suffix in self.LOADERS:
                self.df = self.LOADERS[suffix](event.path)
                self.__load_table()
            else:
                self.sub_title = f"Unsupported file type:  {suffix}"
        except Exception as exc:
            self.sub_title = f"ERROR {exc} {suffix}"
        else:
            # data_view.update(syntax)
            # self.query_one("#data-view").scroll_home(animate=False)
            self.sub_title = event.path

    def action_toggle_files(self) -> None:
        """Called in response to key binding."""
        self.show_tree = not self.show_tree

    def action_toggle_dtype(self) -> None:
        """Called in response to key binding."""
        self.show_dtype = not self.show_dtype
        self.__load_table()

    def action_screenshot(self, filename=None, path: str = "./") -> None:
        """Save an SVG "screenshot". This action will save an SVG file containing the current contents of the screen.

        Args:
            filename: Filename of screenshot, or None to auto-generate. Defaults to None.
            path: Path to directory. Defaults to "./".
        """
        self.bell()
        path = self.save_screenshot(filename, path)
        message = Text.assemble("Screenshot saved to ", (f"'{path}'", "bold green"))
        self.screen.mount(Notification(message))


def run():
    DataBrowser().run()


if __name__ == "__main__":
    run()
